from typing import Any, Dict, Iterable, List, NamedTuple, TextIO, Tuple
from os.path import exists, join as join_path
import numpy as np
from pandas import DataFrame
from ase import Atoms
from ase.units import GPa
from ase.io import read, write
from ase.stress import voigt_6_to_full_3x3_stress


def read_loss(filename: str) -> DataFrame:
    """Parses a file in ``loss.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_output_files_for_the_nep_executable#The_loss.out_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) != 10:
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected 10 columns.')
    tags = 'total_loss L1 L2'
    tags += ' RMSE_E_train RMSE_F_train RMSE_V_train'
    tags += ' RMSE_E_test RMSE_F_test RMSE_V_test'
    generations = range(100, len(data) * 100 + 1, 100)
    df = DataFrame(data=data[:, 1:], columns=tags.split(), index=generations)
    return df


def read_potential(filename: str) -> Tuple[Dict[str, tuple], List[float]]:
    """Parses a file in ``nep.txt`` format from GPUMD and returns the
    content as a tuple of a dict and a list. The dict contains the
    information contained in the header, whereas the list contains the
    weights and rescaling factors.

    Parameters
    ----------
    filename
        input file name

    """
    header = []
    parameters = []
    nheader = 6
    with open(filename) as f:
        for k, line in enumerate(f.readlines()):
            flds = line.split()
            assert len(flds) != 0, f'Empty line number {k}'
            if k == 0 and 'zbl' in flds[0]:
                nheader += 1
            if k <= nheader:
                header.append(tuple(flds))
            elif len(flds) == 1:
                parameters.append(float(flds[0]))
            else:
                raise IOError(f'Failed to parse line {k} from {filename}')
    settings = {}
    for flds in header:
        if flds[0] in ['cutoff']:
            settings[flds[0]] = tuple(map(float, flds[1:]))
        elif flds[0] in ['zbl', 'n_max', 'l_max', 'ANN', 'basis_size']:
            settings[flds[0]] = tuple(map(int, flds[1:]))
        else:
            settings[flds[0]] = flds[1:]
    return settings, parameters


def _write_structure_in_nep_format(structure: Atoms,
                                   f: TextIO) -> None:
    """Write structure block into a file-like object in format readable by nep executable.

    Parameters
    ----------
    structure
        input structure; must hold information regarding energy and forces
    f
        file-like object to which to write
    """

    # Allowed keyword=value pairs. Use ASEs extyz write functionality.:
    #   lattice="ax ay az bx by bz cx cy cz"                    (mandatory)
    #   energy=energy_value                                     (mandatory)
    #   virial="vxx vxy vxz vyx vyy vyz vzx vzy vzz"            (optional)
    #   weight=relative_weight                                  (optional)
    #   properties=property_name:data_type:number_of_columns
    #       species:S:1                      (mandatory)
    #       pos:R:3                          (mandatory)
    #       force:R:3 or forces:R:3          (mandatory)
    try:
        energy = structure.get_potential_energy()
        structure.get_forces()  # calculate forces to have them on the Atoms object
        structure.info['energy'] = energy
    except RuntimeError:
        raise RuntimeError('Failed to retrieve energy and/or forces for structure')
    if np.isclose(structure.get_volume(), 0):
        raise ValueError('Structure cell must have a non-zero volume!')
    try:
        stress = structure.get_stress(voigt=False).flatten()
        stress = ' '.join([f'{s:.6f}' for s in stress])
        structure.info['stress'] = f'\"{stress}\"'
    except RuntimeError:
        print('Warning: Failed to retrieve stresses for structure')
    write(filename=f, images=structure, write_info=True, format='extxyz')


def write_structures(outfile: str,
                     structures: List[Atoms]) -> None:
    """Writes structures for training/testing in format readable by nep executable.

    Parameters
    ----------
    outfile
        output filename
    structures
        list of structures with energy, forces, and (possibly) stresses
    """
    with open(outfile, 'w') as f:
        for structure in structures:
            _write_structure_in_nep_format(structure, f)


def write_nepfile(parameters: NamedTuple,
                  dirname: str) -> None:
    """Writes parameters file for NEP construction.

    Parameters
    ----------
    parameters
        input parameters; see https://gpumd.zheyongfan.org/index.php/The_nep.in_input_file
    dirname
        directory in which to place input file and links
    """
    with open(join_path(dirname, 'nep.in'), 'w') as f:
        for key, val in parameters.items():
            f.write(f'{key}  ')
            if isinstance(val, Iterable):
                f.write(' '.join([f'{v}' for v in val]))
            else:
                f.write(f'{val}')
            f.write('\n')


def read_nepfile(filename: str) -> Dict[str, Any]:
    """Returns the content of a configuration file (`nep.in`) as a dictionary.

    Parameters
    ----------
    filename
        input file name
    """
    settings = {}
    with open(filename) as f:
        for line in f.readlines():
            flds = line.split()
            if len(flds) == 0:
                continue
            if flds[0].startswith('#'):
                continue
            settings[flds[0]] = ' '.join(flds[1:])
    for key, val in settings.items():
        if key in ['version', 'neuron', 'generation', 'batch', 'population']:
            settings[key] = int(val)
        elif key in ['lambda_1', 'lambda_2', 'lambda_e', 'lambda_f', 'lambda_v',
                     'lambda_shear', 'force_delta']:
            settings[key] = float(val)
        elif key in ['cutoff', 'n_max', 'l_max', 'basis_size', 'zbl', 'type_weight']:
            settings[key] = [float(v) for v in val.split()]
        elif key == 'type':
            types = val.split()
            types[0] = int(types[0])
            settings[key] = types
    return settings


def read_structures(dirname: str) -> Tuple[List[Atoms], List[Atoms]]:
    """Parses the energy_*.out, force_*.out, and virial_*.out files from a
    nep run and returns their content as lists.  The first and
    second list contain the structures from the training and
    test sets, respectively. Each list entry corresponds to an
    ASE Atoms object, which in turn contains predicted and target
    energies, forces and virials/stresses stored in the `info`
    property.

    Parameters
    ----------
    dirname
        directory from which to read output files
    """
    path = join_path(dirname)
    if not exists(path):
        raise ValueError(f'Directory {path} does not exist')

    structures = {}

    for stype in ['train', 'test']:
        structures[stype] = read(join_path(dirname, f'{stype}.xyz'), format='extxyz', index=':')
        n_atoms = [len(atom) for atom in structures[stype]]
        cells = [atom.cell for atom in structures[stype]]
        n_structures = len(structures[stype])

        ts, ps = _read_data_file(path, f'energy_{stype}.out')
        assert len(ts) == n_structures, \
            f'Number of structures in energy_{stype}.out ({len(ts)})' \
            f' and {stype}.xyz ({n_structures}) inconsistent'
        for k, t, p in zip(range(n_structures), ts, ps):
            structures[stype][k].info['energy_target'] = t
            structures[stype][k].info['energy_predicted'] = p

        ts, ps = _read_data_file(path, f'force_{stype}.out')
        assert len(ts) == sum(n_atoms), \
            f'Number of structures in force_{stype}.out ({len(ts)})' \
            f' and {stype}.xyz ({n_structures}) inconsistent'
        n = 0
        for k, nat in zip(range(n_structures), n_atoms):
            structures[stype][k].info['force_target'] = np.array(ts[n:n+nat]).reshape(nat, 3)
            structures[stype][k].info['force_predicted'] = np.array(ps[n:n+nat]).reshape(nat, 3)
            n += nat

        ts, ps = _read_data_file(path, f'virial_{stype}.out')
        assert len(ts) == 6 * n_structures, \
            f'Number of structures in virial_{stype}.out ({len(ts)})' \
            f' and {stype}.xyz ({n_structures}) inconsistent'
        ts = np.array(ts).reshape((6, n_structures)).T
        ps = np.array(ps).reshape((6, n_structures)).T
        for k, nat, cell, t, p in zip(range(n_structures), n_atoms, cells, ts, ps):
            assert np.shape(t) == (6,)
            structures[stype][k].info['virial_target'] = t
            structures[stype][k].info['virial_predicted'] = p
            conv = nat / np.linalg.det(cell) / GPa
            structures[stype][k].info['stress_target'] = t * conv
            structures[stype][k].info['stress_predicted'] = p * conv
    return structures['train'], structures['test']


def _read_data_file(dirname: str,
                    fname: str):
    """Private function that parses energy/force/virial_*.out files and
    returns their content for further processing.
    """
    path = join_path(dirname, fname)
    if not exists(path):
        raise ValueError(f'Directory {path} does not exist')
    with open(path, 'r') as f:
        lines = f.readlines()
    target, predicted = [], []
    for line in lines:
        flds = line.split()
        if len(flds) == 6:
            predicted.append([float(s) for s in flds[0:3]])
            target.append([float(s) for s in flds[3:6]])
        elif len(flds) == 2:
            predicted.append(float(flds[0]))
            target.append(float(flds[1]))
    return target, predicted


def get_parity_data(structures: List[Atoms],
                    property: str,
                    selection: List[str] = None,
                    flatten: bool = False) -> DataFrame:
    """Returns the predicted and target energies, forces, virials or stresses
    from a list of structures in a format suitable for generating parity plots.

    The structures should  have been read using :func:`read_structures
    <calorine.nep.io.read_structures>`, such that the ``info``-object is
    populated with keys on the form ``{property}_{type}`` where :attr:`{property}`
    is one of energy, force, virial, and stress, and ``{type}`` is one
    of predicted or target.

    The resulting parity data is returned as a tuple of dicts, where each entry
    corresponds to a list.

    Parameters
    ----------
    structures
        List of structures as read with `read_structures`.
    property
        One of 'energy', 'force', 'virial', 'stress'.
    selection
        A list containing which components to return, and/or the absolute value.
        Possible values are 'x', 'y', 'z', 'xx', 'yy', 'zz', 'yz', 'xz', 'xy', 'abs', 'pressure'.
    flatten
        if True return flattened lists; this is useful for flattening
        the components of force or virials into a simple list
    """
    data = {
        'predicted': [],
        'target': []
    }
    voigt_mapping = {
        'x': 0,
        'y': 1,
        'z': 2,
        'xx': 0,
        'yy': 1,
        'zz': 2,
        'yz': 3,
        'xz': 4,
        'xy': 5
    }
    if property not in ('energy', 'force', 'virial', 'stress'):
        raise ValueError("`property` must be one of 'energy', 'force', 'virial', 'stress'.")
    if property == 'energy' and selection:
        raise ValueError('Selection does nothing for scalar-valued `energy`.')
    if property != 'stress' and selection and 'pressure' in selection:
        raise ValueError(f'Cannot calculate pressure for `{property}`.')
    for structure in structures:
        for stype in data:
            property_with_stype = f'{property}_{stype}'
            if property_with_stype not in structure.info.keys():
                raise KeyError(f'{property_with_stype} does not exist in info object!')
            extracted_property = np.array(structure.info[property_with_stype])

            if selection is None or len(selection) == 0:
                data[stype].append(extracted_property)
                continue

            selected_values = []
            for select in selection:
                if property == 'force':
                    # flip to get (n_components, n_structures)
                    extracted_property = extracted_property.T
                if select == 'abs':
                    if property == 'force':
                        selected_values.append(np.linalg.norm(extracted_property, axis=0))
                    else:
                        # property can only be in ('virial', 'stress')
                        full_tensor = voigt_6_to_full_3x3_stress(extracted_property)
                        selected_values.append(np.linalg.norm(full_tensor))
                    continue

                if select == 'pressure' and property == 'stress':
                    total_stress = extracted_property
                    selected_values.append(- np.sum(total_stress[:3]) / 3)
                    continue

                if select not in voigt_mapping:
                    raise ValueError(f'Selection `{select}` is not allowed.')
                index = voigt_mapping[select]
                if index >= extracted_property.shape[0]:
                    raise ValueError(
                        f'Selection `{select}` is not compatible with property `{property}`.')
                selected_values.append(extracted_property[index])
            data[stype].append(selected_values)
    if flatten:
        for key, value in data.items():
            if len(np.shape(value[0])) > 0:
                data[key] = np.concatenate(value).ravel().tolist()
    return DataFrame(data)



from typing import List

import numpy as np
from ase import Atoms
from ase.io import read, write
from pandas import DataFrame


def read_kappa(filename: str) -> DataFrame:
    """Parses a file in ``kappa.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_kappa.out_output_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in kappa.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'kx_in kx_out ky_in ky_out kz_tot'.split()
    if len(data[0]) != len(tags):
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected {len(tags)} columns.')
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    return df


def read_hac(filename: str) -> DataFrame:
    """Parses a file in ``hac.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_hac.out_output_file.

    Parameters
    ----------
    filename
        input file name

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in hac.out, append a dimension
        data = data.reshape(1, -1)
    tags = 'time'
    tags += ' jin_jtot_x jout_jtot_x jin_jtot_y jout_jtot_y jtot_jtot_z'
    tags += ' kx_in kx_out ky_in ky_out kz_tot'
    tags = tags.split()
    if len(data[0]) != len(tags):
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         f' Expected {len(tags)} columns.')
    df = DataFrame(data=data, columns=tags)
    df['kx_tot'] = df.kx_in + df.kx_out
    df['ky_tot'] = df.ky_in + df.ky_out
    # remove columns with less relevant data to save space
    for col in df:
        if 'jtot' in col or '_in' in col:
            del df[col]
    return df


def read_thermo(filename: str,
                natoms: int = 1) -> DataFrame:
    """Parses a file in ``thermo.out`` format from GPUMD and returns the
    content as a data frame. More information concerning file format,
    content and units can be found at
    https://gpumd.zheyongfan.org/index.php/The_thermo.out_output_file.

    Parameters
    ----------
    filename
        input file name
    natoms
        number of atoms; used to normalize energies

    """
    data = np.loadtxt(filename)
    if isinstance(data[0], np.float64):
        # If only a single row in loss.out, append a dimension
        data = data.reshape(1, -1)
    if len(data[0]) == 9:
        # orthorhombic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 12:
        # orthorhombic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += ' cell_xx cell_yy cell_zz'
    elif len(data[0]) == 15:
        # triclinic box
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz'
        tags += ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
    elif len(data[0]) == 18:
        # triclinic box with stresses in Voigt notation (v3.3.1+)
        tags = 'temperature kinetic_energy potential_energy'
        tags += ' stress_xx stress_yy stress_zz stress_yz stress_xz stress_xy'
        tags += ' cell_xx cell_xy cell_xz cell_yx cell_yy cell_yz cell_zx cell_zy cell_zz'
    else:
        raise ValueError(f'Input file contains {len(data[0])} data columns.'
                         ' Expected 9, 12, 15 or 18 columns.')
    df = DataFrame(data=data, columns=tags.split())
    assert natoms > 0, 'natoms must be positive'
    df.kinetic_energy /= natoms
    df.potential_energy /= natoms
    return df


def read_xyz(filename: str) -> Atoms:
    """
    Read the structure input file (`model.xyz`) for GPUMD and return the
    structure along with run input parameters from the file.

    This is a wrapper function around :func:`ase.io.read_xyz` since the ASE implementation does
    not read velocities properly.

    Parameters
    ----------
    filename
        Name of file from which to read the structure

    Returns
    -------
    structure as ASE Atoms object with additional per-atom arrays
    representing atomic masses, velocities etc.
    """
    structure = read(filename, format='extxyz')
    if structure.has('vel'):
        structure.set_velocities(structure.get_array('vel'))
    return structure


def write_xyz(filename: str,
              structure: Atoms,
              groupings: List[List[List[int]]] = None):
    """
    Writes a structure into GPUMD input format (`model.xyz`).

    Parameters
    ----------
    filename
        Name of file to which the structure should be written
    structure
        Input structure
    groupings
        Groups into which the individual atoms should be divided in the form of
        a list of list of lists. Specifically, the outer list corresponds to
        the grouping methods, of which there can be three at the most, which
        contains a list of groups in the form of lists of site indices. The
        sum of the lengths of the latter must be the same as the total number
        of atoms.

    Raises
    ------
    ValueError
        Raised if parameters are incompatible
    """
    # Make a local copy of the atoms object
    _structure = structure.copy()

    # Check velocties parameter
    velocities = _structure.get_velocities()
    if velocities is None or np.max(np.abs(velocities)) < 1e-6:
        has_velocity = 0
    else:
        has_velocity = 1

    # Check groupings parameter
    if groupings is None:
        number_of_grouping_methods = 0
    else:
        number_of_grouping_methods = len(groupings)
        if number_of_grouping_methods > 3:
            raise ValueError('There can be no more than 3 grouping methods!')
        for g, grouping in enumerate(groupings):
            all_indices = [i for group in grouping for i in group]
            if len(all_indices) != len(_structure) or \
                    set(all_indices) != set(range(len(_structure))):
                raise ValueError(f'The indices listed in grouping method {g} are'
                                 ' not compatible with the input'
                                 ' structure!')

    # Allowed keyword=value pairs. Use ASEs extyz write functionality.:
    #   pbc="pbc_a pbc_b pbc_c"
    #   lattice="ax ay az bx by bz cx cy cz"
    #   properties=property_name:data_type:number_of_columns
    #       species:S:1
    #       pos:R:3
    #       mass:R:1
    #       vel:R:3
    #       group:I:number_of_grouping_methods
    _structure.new_array('mass', _structure.get_masses())
    if has_velocity:
        _structure.new_array('vel', _structure.get_velocities())
    if groupings is not None:
        group_indices = np.array([
            [
                [group_index for group_index, group in enumerate(
                    grouping) if structure_idx in group] for grouping in groupings
            ]
            for structure_idx in range(len(_structure))]).squeeze()  # pythoniccc
        _structure.new_array('group', group_indices)

    write(filename=filename, images=_structure, write_info=True, format='extxyz')



# -*- coding: utf-8 -*-
"""
Main module of the calorine package.
"""

__project__ = 'calorine'
__description__ = 'A Python library for building and sampling NEP models via GPUMD simulations.'
__copyright__ = '2023'
__license__ = 'Mozilla Public License 2.0 (MPL 2.0)'
__version__ = '1.4'
__maintainer__ = 'The calorine developers team'
__status__ = 'Stable'
__url__ = 'http://calorine.materialsmodeling.org/'



# -*- coding: utf-8 -*-
from .io import (read_loss,
                 read_nepfile,
                 read_potential,
                 read_structures,
                 write_nepfile,
                 write_structures,
                 get_parity_data)
from .nep import (get_descriptors,
                  get_potential_forces_and_virials)
from .training_factory import setup_training

__all__ = ['read_loss',
           'read_nepfile',
           'read_potential',
           'read_structures',
           'get_parity_data',
           'get_descriptors',
           'get_potential_forces_and_virials',
           'setup_training',
           'write_nepfile',
           'write_structures']



import pytest
import numpy as np
from ase.io import read
from ase import Atoms
from ase.build import bulk
from ase.stress import voigt_6_to_full_3x3_stress
from ase.calculators.lj import LennardJones
from calorine.nep import (read_loss, read_nepfile, read_potential,
                          write_structures, get_parity_data)


@pytest.fixture
def PbTe():
    PbTeBulk = bulk('PbTe', crystalstructure='rocksalt', a=4)
    PbTeBulk[0].position += np.array([0.03, 0.02, 0])
    return PbTeBulk


@pytest.fixture
def TenPbTeWithInfo(PbTe):
    n_structures = 10
    structures = []
    for i in range(n_structures):
        # Test with differently sized structures in the test set
        atoms = PbTe.copy().repeat(i+1)
        natoms = len(atoms)
        atoms.info['energy_target'] = float(i)
        atoms.info['energy_predicted'] = float(i)
        atoms.info['force_target'] = np.arange(i, natoms*3+i).reshape((natoms, 3))
        atoms.info['force_predicted'] = np.arange(i, natoms*3+i).reshape((natoms, 3))
        atoms.info['virial_target'] = -np.arange(i, 6+i).reshape(6)
        atoms.info['virial_predicted'] = -np.arange(i, 6+i).reshape(6)
        atoms.info['stress_target'] = -np.arange(i, 6+i).reshape(6)
        atoms.info['stress_predicted'] = -np.arange(i, 6+i).reshape(6)
        structures.append(atoms)
    return structures


def sel2idx(select: str) -> int:
    map = {
        'x': 0,
        'y': 1,
        'z': 2,
        'xx': 0,
        'yy': 1,
        'zz': 2,
        'yz': 3,
        'xz': 4,
        'xy': 5
    }
    return map[select]


# --- read_nepfile ---
def test_read_nepfile_unknown_setting(tmpdir):
    """Reads a nep.in file with a comment"""
    p = tmpdir.join('nep.in')
    p.write('test 1\n')
    settings = read_nepfile(str(p))
    assert settings['test'] == '1'


def test_read_nepfile_run_comment(tmpdir):
    """Reads a nep.in file with a comment"""
    p = tmpdir.join('nep.in')
    p.write('#basis_size 10 6\n')
    settings = read_nepfile(str(p))
    assert settings == {}


def test_read_nepfile_nep():
    """Reads a nep.in file"""
    settings = read_nepfile('tests/example_files/nep.in')
    assert settings['version'] == 4
    assert settings['type'] == [2, 'C', 'H']
    assert settings['cutoff'] == [8, 4]
    assert settings['n_max'] == [8, 6]
    assert settings['l_max'] == [4]
    assert settings['neuron'] == 50
    assert settings['lambda_1'] == 0.01
    assert settings['lambda_2'] == 0.01
    assert settings['batch'] == 100
    assert settings['population'] == 50
    assert settings['generation'] == 50000
    assert settings['lambda_e'] == 1
    assert 'lambda_f' not in settings.keys()


def test_read_nepfile_nep_blank_line(tmpdir):
    """Reads a nep.in file with a blank line"""
    p = tmpdir.join('nep.in')
    p.write('\n')
    settings = read_nepfile(str(p))
    assert settings == {}


# --- read_loss ---
def test_read_loss():
    """Reads a loss.out file"""
    loss = read_loss('tests/example_files/loss.out')
    columns_check = loss.columns == ['total_loss', 'L1', 'L2',
                                     'RMSE_E_train', 'RMSE_F_train', 'RMSE_V_train',
                                     'RMSE_E_test', 'RMSE_F_test', 'RMSE_V_test']
    assert columns_check.all()
    assert isinstance(loss.index[0], int)
    assert loss.index[0] == 100
    assert len(loss) == 95


def test_read_loss_single_row(tmpdir):
    """Tries to read a loss.out file that has only a single row"""
    p = tmpdir.join('loss.out')
    p.write('100 2 3 4 5 6 7 8 9 10\n')
    loss = read_loss(str(p))
    columns_check = loss.columns == ['total_loss', 'L1', 'L2',
                                     'RMSE_E_train', 'RMSE_F_train', 'RMSE_V_train',
                                     'RMSE_E_test', 'RMSE_F_test', 'RMSE_V_test']
    assert columns_check.all()
    assert isinstance(loss.index[0], int)
    assert loss.index[0] == 100


def test_read_loss_malformed_file(tmpdir):
    """Tries to read a malformed loss.out file"""
    p = tmpdir.join('loss_invalid.out')
    p.write('0 0 0 0 0 0\n')
    with pytest.raises(ValueError) as e:
        read_loss(str(p))
    assert 'Input file contains 6 data columns. Expected 10 columns.' in str(e)


# --- read_potential ---
def test_read_potential():
    """Reads a NEP3 nep.txt file"""
    settings, parameters = read_potential('tests/example_files/nep.txt')
    assert settings['zbl'] == (1, 2)
    assert settings['cutoff'] == (8, 4)
    assert settings['n_max'] == (4, 4)
    assert settings['l_max'] == (4, 2, 0)
    assert settings['basis_size'] == (8, 8)
    assert settings['ANN'] == (30, 0)
    assert len(parameters) == 1350


def test_read_potential_malformed_file(tmpdir):
    """Tries to read an invalid nep.in file"""
    p = tmpdir.join('nep.in')
    p.write("""nep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\nnep123 x x\n
            nep123 x x\nnep123 x x\n  6.2799901e-02""")
    with pytest.raises(ValueError) as e:
        read_loss(str(p))
    assert 'could not convert string' in str(e)


# --- write_structures ---
def test_write_structures_single_structure(tmpdir, PbTe):
    """Writes a structure to a extxyz-file"""
    p = tmpdir.join('train.xyz')
    PbTe.calc = LennardJones()
    write_structures(str(p), [PbTe])
    read_structure = read(str(p), format='extxyz')
    assert np.isclose(read_structure.cell.volume, PbTe.cell.volume, atol=1e-12, rtol=1e-6)
    assert np.isclose(read_structure.get_potential_energy(),
                      PbTe.get_potential_energy(), atol=1e-12, rtol=1e-6)
    assert np.allclose(read_structure.positions, PbTe.positions, atol=1e-12, rtol=1e-6)
    assert np.allclose(read_structure.get_forces(), PbTe.get_forces(), atol=1e-12, rtol=1e-6)
    # Written accuracy is around 2e-6
    assert np.allclose(read_structure.get_stress(),
                       PbTe.get_stress(voigt=True), atol=1e-12, rtol=1e-5)


def test_write_structures_with_weight(tmpdir, PbTe):
    """Writes structures with weight to an extxyz-file"""
    p = tmpdir.join('train.xyz')
    structure1 = PbTe.copy()
    structure2 = PbTe.copy()
    structure1.calc = LennardJones()
    structure2.calc = LennardJones()
    structure1.info['weight'] = 1
    structure2.info['weight'] = 50
    write_structures(str(p), [structure1, structure2])
    with open(str(p), 'r') as f:
        lines = f.readlines()
        assert 'weight=1' in lines[1]
        assert 'weight=50' in lines[5]


def test_write_structure_with_filename(tmpdir, PbTe):
    """Writes with a filename to an extxyz-file"""
    p = tmpdir.join('train.xyz')
    PbTe.calc = LennardJones()
    PbTe.info['filename'] = 'result.testcar'
    write_structures(str(p), [PbTe])
    read_structure = read(str(p), format='extxyz')
    print(read_structure.info)
    assert read_structure.info['filename'] == 'result.testcar'
    with open(str(p), 'r') as f:
        lines = f.readlines()
        assert 'filename=result.testcar' in lines[1]


def test_write_structure_without_energy_or_forces(tmpdir):
    """Tries to write a structure without a calculator attached"""
    p = tmpdir.join('train.xyz')
    C = Atoms('C', positions=[(0, 0, 0)])
    with pytest.raises(RuntimeError) as e:
        write_structures(str(p), [C])
    assert 'Failed to retrieve energy and/or forces for structure' in str(e)


def test_write_structure_without_cell(tmpdir):
    """Tries to write a structure without a calculator attached"""
    p = tmpdir.join('train.xyz')
    C = Atoms('C', positions=[(0, 0, 0)])
    C.calc = LennardJones()
    with pytest.raises(ValueError) as e:
        write_structures(str(p), [C])
    assert 'You have 0 lattice vectors: volume not defined' in str(e)


# get_parity_data
@pytest.mark.parametrize('property',
                         ['energy', 'force', 'virial', 'stress'])
def test_get_parity_data(property, TenPbTeWithInfo):
    """Extracts parity data from a list of structures"""
    df = get_parity_data(structures=TenPbTeWithInfo, property=property)
    assert np.all(df.columns == ['predicted', 'target'])
    assert len(df['target']) == 10
    assert len(df['target']) == len(df['predicted'])
    if not property == 'energy':
        assert df['target'][0].shape == df['predicted'][0].shape


@pytest.mark.parametrize('property, selection',
                         [('force', ['x']),
                          ('virial', ['x', 'y', 'z']),
                          ('stress', ['xx', 'yy', 'zz'])
                          ])
def test_get_parity_data_diagonal_components(TenPbTeWithInfo, property, selection):
    """Extracts parity data from a list of structures, selecting properties on the diagonal."""
    df = get_parity_data(structures=TenPbTeWithInfo, property=property, selection=selection)
    for i, structure in enumerate(TenPbTeWithInfo):
        assert len(df['target'][i]) == len(selection)
        assert len(df['target'][i]) == len(df['predicted'][i])
        for j, select in enumerate(selection):
            if property == 'force':
                expected_target = structure.info[f'{property}_target'][:, sel2idx(select)]
                expected_predicted = structure.info[f'{property}_predicted'][:, sel2idx(select)]
            elif property in ('virial', 'stress'):
                expected_target = structure.info[f'{property}_target'][sel2idx(select)]
                expected_predicted = structure.info[f'{property}_predicted'][sel2idx(select)]
            assert np.all(df['target'][i][j] == expected_target)
            assert np.all(df['predicted'][i][j] == expected_predicted)


@pytest.mark.parametrize('property, selection',
                         [('virial', ['yz', 'xz', 'xy']),
                          ('stress', ['yz', 'xz', 'xy'])
                          ])
def test_get_parity_data_off_diagonal_components(TenPbTeWithInfo, property, selection):
    """Extracts parity data from a list of structures, selecting properties on the off-diagonal."""
    df = get_parity_data(structures=TenPbTeWithInfo, property=property, selection=selection)
    for i, structure in enumerate(TenPbTeWithInfo):
        assert len(df['target'][i]) == len(selection)
        assert len(df['target'][i]) == len(df['predicted'][i])
        for j, select in enumerate(selection):
            expected_target = structure.info[f'{property}_target'][sel2idx(select)]
            expected_predicted = structure.info[f'{property}_predicted'][sel2idx(select)]
            assert np.all(df['target'][i][j] == expected_target)
            assert np.all(df['predicted'][i][j] == expected_predicted)


@pytest.mark.parametrize('property, selection',
                         [('force', ['abs']),
                          ('virial', ['abs']),
                          ('stress', ['abs'])
                          ])
def test_get_parity_data_abs(TenPbTeWithInfo, property, selection):
    """Extracts parity data from a list of structures, calculating the absolute value"""
    df = get_parity_data(structures=TenPbTeWithInfo, property=property, selection=selection)
    for i, structure in enumerate(TenPbTeWithInfo):
        assert len(df['target'][i]) == len(selection)
        assert len(df['target'][i]) == len(df['predicted'][i])
        for j, _ in enumerate(selection):
            if property == 'force':
                expected_target = np.linalg.norm(structure.info[f'{property}_target'], axis=1)
                expected_predicted = np.linalg.norm(structure.info[f'{property}_predicted'], axis=1)
            elif property in ('virial', 'stress'):
                expected_target = np.linalg.norm(voigt_6_to_full_3x3_stress(
                    structure.info[f'{property}_target']))
                expected_predicted = np.linalg.norm(voigt_6_to_full_3x3_stress(
                    structure.info[f'{property}_predicted']))
            assert np.all(df['target'][i][j] == expected_target)
            assert np.all(df['predicted'][i][j] == expected_predicted)


@pytest.mark.parametrize('property, selection',
                         [('stress', ['pressure'])])
def test_get_parity_data_pressure(TenPbTeWithInfo, property, selection):
    """Extracts parity data from a list of structures, calculating the pressure"""
    df = get_parity_data(structures=TenPbTeWithInfo, property=property, selection=selection)
    for i, structure in enumerate(TenPbTeWithInfo):
        assert len(df['target'][i]) == len(selection)
        assert len(df['target'][i]) == len(df['predicted'][i])
        for j, _ in enumerate(selection):
            expected_target = -np.sum(structure.info[f'{property}_target'][:3]) / 3
            expected_predicted = -np.sum(structure.info[f'{property}_predicted'][:3]) / 3
            assert np.all(df['target'][i][j] == expected_target)
            assert np.all(df['predicted'][i][j] == expected_predicted)


@pytest.mark.parametrize('property, selection, output',
                         [('energy', ['x'],
                           'Selection does nothing for scalar-valued `energy`.'),
                          ('force', ['pressure'],
                          'Cannot calculate pressure for `force`.'),
                          ('force', ['xy'],
                          'Selection `xy` is not compatible with property `force`.'),
                          ('virial', ['zy'],
                          'Selection `zy` is not allowed.')
                          ])
def test_get_parity_data_invalid_selection(TenPbTeWithInfo, property, selection, output):
    """Tries to extract parity data from structures with an invalid selection"""
    with pytest.raises(ValueError) as e:
        get_parity_data(structures=TenPbTeWithInfo, property=property, selection=selection)
    assert output in str(e)


@pytest.mark.parametrize('property, output',
                         [('bounciness',
                           "`property` must be one of 'energy', 'force', 'virial', 'stress'."),
                          ])
def test_get_parity_data_invalid_property(TenPbTeWithInfo, property, output):
    """Tries to extract parity data from structures with an invalid property"""
    with pytest.raises(ValueError) as e:
        get_parity_data(structures=TenPbTeWithInfo, property=property)
    assert output in str(e)


def test_get_parity_data_missing_info():
    """Raises error when property(ies) is missing in atom info dictionary"""
    n_structures = 10
    structures = []
    for _ in range(n_structures):
        # Test with differently sized structures in the test set
        natoms = np.random.randint(1, 10)
        atoms = Atoms('C'*natoms)
        atoms.info['energy_target'] = 1.0
        structures.append(atoms)
    with pytest.raises(KeyError) as e:
        get_parity_data(structures, property='energy')
    assert 'energy_predicted does not exist in info object!' in str(e)


@pytest.mark.parametrize('property, count_wo_flatten, count_with_flatten',
                         [('energy', 10, 10),
                          ('force', 10, 18150),
                          ('virial', 10, 60),
                          ])
def test_get_parity_data_flatten(property, count_wo_flatten, count_with_flatten, TenPbTeWithInfo):
    """Checks the option to flatten the data."""
    df_wo_flatten = get_parity_data(structures=TenPbTeWithInfo, property=property, flatten=False)
    assert len(df_wo_flatten) == count_wo_flatten
    df_with_flatten = get_parity_data(structures=TenPbTeWithInfo, property=property, flatten=True)
    assert len(df_with_flatten) == count_with_flatten



import tempfile
import subprocess
import shutil


executable = '/home/elindgren/repos/NEP_CPU/src/a.out'
structure_dir = '../example_structures'
output_dir = '../example_output/'


def execute_model_in_temporary_directory(model, structure):
    with tempfile.TemporaryDirectory() as tmpdirname:
        print('created temporary directory', tmpdirname)
        # copy structure and model
        shutil.copy(f'{model}.txt', f'{tmpdirname}/nep.txt')
        shutil.copy(f'{structure_dir}/{structure}.in', f'{tmpdirname}/xyz.in')
        # Execute
        subprocess.Popen(["cat", "xyz.in"],  cwd=tmpdirname)
        subprocess.Popen([executable, '.'], cwd=tmpdirname).wait()
        # Copy back results
        shutil.copy(f'{tmpdirname}/descriptor.out',
                    f'{output_dir}/{model}_{structure}_descriptor.out')
        shutil.copy(f'{tmpdirname}/force_analytical.out',
                    f'{output_dir}/{model}_{structure}_force.out')
        shutil.copy(f'{tmpdirname}/virial.out',
                    f'{output_dir}/{model}_{structure}_virial.out')


models_and_structures = [
    ('C_NEP2_dummy', 'C_2atom'),
    ('CO_NEP2_dummy', 'CO_2atom'),
    ('CON_NEP2_dummy', 'CON_3atom'),
    ('PbTe_NEP2_dummy', 'PbTe_2atom'),
    ('nep3_v3.3.1_PbTe_Fan22', 'PbTe_2atom'),
    ('nep3_v3.3.1_PbTe_Fan22', 'PbTe_250atom'),
]


for model, structure in models_and_structures:
    execute_model_in_temporary_directory(model, structure)



import pytest
import numpy as np
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.io import read
from os.path import exists, join as join_path
from calorine.nep import read_nepfile, setup_training


@pytest.fixture
def nep_parameters():
    return dict(version=3,
                type=[2, 'Au', 'Cu'],
                cutoff=[8, 4],
                n_max=[8, 6],
                l_max=[4, 0],
                lambda_1=0.1,
                lambda_2=0.2,
                lambda_e=1,
                lambda_f=5,
                lambda_v=0.2,
                neuron=50,
                generation=200000)


@pytest.fixture
def structures():
    structures = []
    for alat in np.arange(3, 6, 0.1):
        structure = bulk('AuCu', crystalstructure='rocksalt', a=alat)
        structure.calc = EMT()
        structures.append(structure)
    return structures


@pytest.fixture
def setup_parameters():
    return dict(rootdir='test_dir',
                train_fraction=0.8,
                seed=42,
                n_splits=3,
                overwrite=True)


def test_setup_training(nep_parameters, structures, setup_parameters):
    setup_training(nep_parameters, structures, **setup_parameters)
    rootdir = setup_parameters['rootdir']
    n_splits = setup_parameters['n_splits']
    train_fraction = setup_parameters['train_fraction']

    assert exists(rootdir), 'Root directory not created'

    path = join_path(rootdir, 'nepmodel_full')
    assert exists(path), 'nepmodel_full subdirectory not created'
    assert exists(join_path(path, 'train.xyz')), 'Training structures not written'
    assert exists(join_path(path, 'test.xyz')), 'Testing structures not written'
    assert exists(join_path(path, 'nep.in')), 'nep.in not written'

    train_structures = read(join_path(path, 'train.xyz'), ':')
    test_structures = read(join_path(path, 'test.xyz'), ':')
    assert len(train_structures) == len(structures), \
        'Incorrect number of training structures generated'
    assert len(test_structures) == 1, \
        'Incorrect number of test structures generated'

    for k in range(1, n_splits + 1):
        path = join_path(rootdir, f'nepmodel_split{k}')
        assert exists(path), f'nepmodel_split{k} subdirectory not created'
        assert exists(join_path(path, 'train.xyz')), 'Training structures not written'
        assert exists(join_path(path, 'test.xyz')), 'Testing structures not written'
        assert exists(join_path(path, 'nep.in')), 'nep.in not written'

        train_structures = read(join_path(path, 'train.xyz'), ':')
        test_structures = read(join_path(path, 'test.xyz'), ':')
        train_size = int(train_fraction * len(structures))
        assert len(train_structures) == train_size, \
            'Incorrect number of training structures generated'
        assert len(test_structures) == len(structures) - train_size, \
            'Incorrect number of test structures generated'


def test_setup_training_invalid_train_fraction(nep_parameters, structures, setup_parameters):
    setup_parameters['train_fraction'] = 1.5
    with pytest.raises(ValueError) as excinfo:
        setup_training(nep_parameters, structures, **setup_parameters)
    assert 'train_fraction (1.5) must fall in the interval (0,1]' in str(excinfo.value), \
        'Incorrect error message for train_fraction outside of interval'


def test_setup_training_existing_dir_no_overwrite(nep_parameters, structures, setup_parameters):
    setup_parameters['overwrite'] = False
    with pytest.raises(FileExistsError) as excinfo:
        setup_training(nep_parameters, structures, **setup_parameters)
    assert 'Output directory exists. Set overwrite=True in' \
        ' order to override this behavior.' in str(excinfo.value), \
        'Incorrect error message for existing directory with no overwrite'


def test_setup_training_invalid_n_splits(nep_parameters, structures, setup_parameters):
    setup_parameters['n_splits'] = 20
    with pytest.raises(ValueError) as excinfo:
        setup_training(nep_parameters, structures, **setup_parameters)
    assert 'n_splits (20) must be positive and must not exceed' in str(excinfo.value), \
        'Incorrect error message for invalid value for n_splits'


def test_setup_training_n_splits_is_none(nep_parameters, structures, setup_parameters):
    setup_parameters['n_splits'] = None
    setup_training(nep_parameters, structures, **setup_parameters)
    path = join_path(setup_parameters['rootdir'], 'nepmodel_full')
    train_structures = read(join_path(path, 'train.xyz'), ':')
    test_structures = read(join_path(path, 'test.xyz'), ':')
    assert len(train_structures) == len(structures), \
        'Incorrect number of training structures generated'
    assert len(test_structures) == 1, \
        'Incorrect number of test structures generated'


def test_setup_training_nep_parameters(nep_parameters, structures, setup_parameters):
    setup_parameters['n_splits'] = None
    setup_training(nep_parameters, structures, **setup_parameters)
    path = join_path(setup_parameters['rootdir'], 'nepmodel_full')
    rec = read_nepfile(join_path(path, 'nep.in'))
    for key, val_ref in nep_parameters.items():
        assert key in rec, f'{key} missing from nep.in'
        val_read = rec[key]
        assert np.all(val_read == val_ref), f'{key} incorrectly written'



from ase import units
from typing import Optional

import numpy as np
import pandas as pd
from scipy import stats


def analyze_data(data: np.ndarray, max_lag: int = None) -> dict:
    """ Carries out an extensive analysis of the data series.

    Parameters
    ----------
    data
        data series to compute autocorrelation function for
    max_lag
        maximum lag between two data points, used for computing autocorrelation

    Returns
    -------
    dict
        calculated properties of the data including, mean, standard deviation,
        correlation length and a 95% error estimate.
    """
    summary = dict(mean=data.mean(),
                   std=data.std())
    acf = get_autocorrelation_function(data, max_lag)
    correlation_length = _estimate_correlation_length_from_acf(acf)
    if correlation_length is not None:
        error_estimate = _estimate_error(data, correlation_length, confidence=0.95)
        summary['correlation_length'] = correlation_length
        summary['error_estimate'] = error_estimate
    else:
        summary['correlation_length'] = np.nan
        summary['error_estimate'] = np.nan
    return summary


def get_autocorrelation_function(data: np.ndarray, max_lag: int = None) -> np.ndarray:
    """ Returns autocorrelation function.

    The autocorrelation function is computed using `pandas.Series.autocorr
    <https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.Series.autocorr.html>`_.

    Parameters
    ----------
    data
        data series to compute autocorrelation function for
    max_lag
        maximum lag between two data points

    Returns
    -------
        calculated autocorrelation function
    """
    if max_lag is None:
        max_lag = len(data) - 1
    if max_lag < 1 or max_lag >= len(data):
        raise ValueError('max_lag should be between 1 and len(data)-1.')
    series = pd.Series(data)
    acf = [series.autocorr(lag) for lag in range(0, max_lag)]
    return np.array(acf)


def get_correlation_length(data: np.ndarray) -> Optional[int]:
    """ Returns estimate of the correlation length of data.

    The correlation length is taken as the first point where the
    autocorrelation functions is less than :math:`\\exp(-2)`. If the
    correlation function never drops below :math:`\\exp(-2)` ``np.nan`` is
    returned.

    If the correlation length cannot be computed since the auto-correlation
    function is unconverged the function returns `None`.

    Parameters
    ----------
    data
        data series for which to the compute autocorrelation function

    Returns
    -------
        correlation length
    """

    acf = get_autocorrelation_function(data)
    correlation_length = _estimate_correlation_length_from_acf(acf)
    if correlation_length is None:
        return None
    return correlation_length


def get_error_estimate(data: np.ndarray, confidence: float = 0.95) -> Optional[float]:
    """Returns estimate of standard error :math:`\\mathrm{error}`
    with confidence interval.

    .. math::

       \\mathrm{error} = t_\\mathrm{factor} * \\mathrm{std}(\\mathrm{data}) / \\sqrt{N_s}

    where :math:`t_{factor}` is the factor corresponding to the confidence
    interval and :math:`N_s` is the number of independent measurements
    (with correlation taken into account).

    If the correlation length cannot be computed since the auto-correlation
    function is unconverged the function returns `None`.

    Parameters
    ----------
    data
        data series for which to estimate the error

    Returns
    -------
        error estimate

    """
    correlation_length = get_correlation_length(data)
    if correlation_length is None:
        return None
    error_estimate = _estimate_error(data, correlation_length, confidence)
    return error_estimate


def _estimate_correlation_length_from_acf(acf: np.ndarray) -> Optional[int]:
    """Estimates correlation length from acf. Returns None if the auto-correlation
    function is uncoverged."""
    for i, a in enumerate(acf):
        if a < np.exp(-2):
            return i
    return None  # np.nan


def _estimate_error(data: np.ndarray,
                    correlation_length: int,
                    confidence: float) -> float:
    """ Estimates error using correlation length. """
    t_factor = stats.t.ppf((1 + confidence) / 2, len(data)-1)  # type: float
    error = t_factor * np.std(data) / np.sqrt(len(data) / correlation_length)  # type: float
    return error


def hac_to_rtc(hac: np.ndarray, V: float, T: float, dt: float) -> np.ndarray:
    """Converts a Heat Auto-Correlation (HAC) to the Running Thermal Conductivity (RTC)

    Parameters
    ----------
    hac
        The HAC <j(t)j(0)> in units of (eV^3 / atomic mass unit) as given by GPUMD
    V
        Volume of cell in Å^3
    T
        Temperature in Kelvin
    dt
        Time step in the HAC in ps

    Returns
    -------
    rtc
        the Running thermal conductivity in (W/m/K)

    Example
    -------
    The general expectation is provided that the hac and kappa are sampled in the same way::
        from calorine.io import read_hac, read_kappa
        hac = read_hac('hac.out').kx_tot
        rtc = read_kappa('kappa.out').kx_tot
        from calorine.tools import hac_to_rtc
        rtc_from_hac = hac_to_rtc(hac, V, T, dt)
        assert np.allclose(rtc, rtc_from_hac)

    """

    hac = np.asarray(hac)

    if hac.ndim > 1:
        raise ValueError('hac must be 1D')

    kB = units.kB  # eV/K

    J_per_eV = 1 / units.J
    kg_per_u = 1 / units.kg
    s_per_ps = 1e-12
    m_per_Å = 1 / units.m

    # The integration and normalization
    rtc = np.cumsum(hac) * dt / (kB * T**2 * V)

    # From eV/ps/Å/K to J/s/m/K
    rtc *= (J_per_eV**3 / kg_per_u) * s_per_ps / (J_per_eV * m_per_Å**3)

    return rtc



import os
from typing import List, Tuple

import numpy as np
import pytest
from ase import Atoms
from ase.build import bulk
from calorine.calculators import GPUNEP


nep2_fname = 'tests/nep_models/PbTe_NEP2_dummy.txt'
nep3_fname = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'


@pytest.fixture
def single_point_parameters() -> List[Tuple]:
    """MD parameters to calculate single point potential energy."""
    return [('dump_thermo', 1),
            ('dump_force', 1),
            ('dump_position', 1),
            ('velocity', 1e-24),
            ('time_step', 1e-6),
            ('ensemble', ['nvt_lan', 1e-24, 1e-24, 1000]),
            ('run', 1)]


@pytest.fixture
def custom_parameters() -> List[Tuple]:
    """Custom MD parameters."""
    return [('dump_thermo', 10),
            ('dump_position', 10),
            ('dump_force', 10),
            ('time_step', 1),
            ('ensemble', ['nvt_ber', 300, 300, 100]),
            ('run', 30)]


@pytest.fixture
def calculator(request, tmpdir) -> GPUNEP:
    """
    Calculator with either a specified directory (tmpdir) or
    automatically created with GPUMD.
    """
    nep_version, specified_dir = request.param
    if nep_version == 3:
        fname = nep3_fname
    elif nep_version == 2:
        fname = nep2_fname
    else:
        raise NotImplementedError('Only NEP2 and NEP3 are currently tested.')
    if specified_dir:
        return GPUNEP(fname,
                      directory=tmpdir)
    else:
        return GPUNEP(fname)


@pytest.fixture
def PbTe() -> Atoms:
    """ASE Atoms object compatible with calculator."""
    structure = bulk('PbTe', crystalstructure='rocksalt', a=4)
    structure[0].position += np.array([0.03, 0.02, 0])
    return structure


def runfile_matches_parameters(fname, parameters, potential_filename):
    """Checks whether content of runfile matches parameters list."""
    with open(fname, 'r') as f:
        for line in f:
            row = line.split()
            found_match = False
            for param in parameters:
                if row[0] == 'potential':
                    if potential_filename not in line:
                        return False
                    found_match = True
                elif row[0] == param[0]:
                    # Convert parameters to iterable
                    if isinstance(param[1], str):
                        param_iter = [param[1]]
                    else:
                        try:
                            iter(param[1])
                            param_iter = param[1]
                        except TypeError:
                            param_iter = [param[1]]

                    # Compare all parameters
                    if len(row[1:]) == len(param_iter):
                        for i, j in zip(row[1:], param_iter):
                            try:
                                i = float(i)
                                if abs(i - j) > 1e-6:
                                    break
                            except ValueError:
                                if i != j:
                                    break
                        else:
                            found_match = True
                            break

            if not found_match:
                return False
    return True


@pytest.mark.parametrize('fname, species',
                         [(nep3_fname, ('Te', 'Pb')),
                          (nep2_fname, ('Te', 'Pb'))])
def test_init(fname, species):
    """Test initialization of calculator."""
    calc = GPUNEP(fname)
    assert isinstance(calc, GPUNEP)
    assert calc.command == 'gpumd > /dev/null'
    assert calc.potential_filename == fname
    assert calc._use_temporary_directory
    assert tuple(calc.species) == species


@pytest.mark.parametrize('fname, directory, label, command',
                         [(nep3_fname, 'tmp',
                           'custom_label', 'custom_command'),
                          (nep2_fname, 'tmp',
                           'custom_label', 'custom_command'),
                          (nep3_fname, 'tmp',
                           'custom_label', 'custom_command > out')])
def test_init_with_custom_settings(fname, directory, label, command):
    """Test initialization with custom settings."""
    calc = GPUNEP(fname,
                  directory=directory,
                  label=label,
                  command=command,
                  )
    if '>' in command:
        assert calc.command == command
    else:
        assert calc.command == command + ' > stdout'
    assert calc.potential_filename == fname
    assert not calc._use_temporary_directory


def test_init_with_nonexistent_model():
    """Test initialization when model file does not exist."""
    with pytest.raises(FileNotFoundError) as e:
        GPUNEP('no-such-model.txt')
    assert 'does not exist' in str(e)


@pytest.mark.parametrize('calculator',
                         [(3, True),
                          (2, True),
                          (3, False)], indirect=['calculator'],
                         )
def test_run_custom_md(PbTe, calculator, custom_parameters):
    """Tests run custom MD function in temporary directory."""
    PbTe.calc = calculator
    atoms = calculator.run_custom_md(custom_parameters, return_last_atoms=True)
    assert isinstance(atoms, Atoms)
    assert not np.allclose(atoms.get_positions(), PbTe.get_positions())

    if not calculator._use_temporary_directory:
        # Check that run file was properly written
        assert runfile_matches_parameters(os.path.join(calculator.directory, 'run.in'),
                                          custom_parameters,
                                          calculator.potential_filename)
    else:
        assert not os.path.exists(calculator.directory)


@pytest.mark.parametrize('calculator',
                         [(3, True)], indirect=['calculator'],
                         )
def test_run_custom_md_only_prepare(PbTe, calculator, custom_parameters):
    """Tests run custom MD function in temporary directory."""
    PbTe.calc = calculator
    retval = calculator.run_custom_md(custom_parameters, only_prepare=True)
    assert retval is None

    # Check that run file was properly written
    assert runfile_matches_parameters(os.path.join(calculator.directory, 'run.in'),
                                      custom_parameters,
                                      calculator.potential_filename)


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_run_custom_md_only_prepare_in_temporary_dir(PbTe, calculator, custom_parameters):
    PbTe.calc = calculator
    with pytest.raises(ValueError) as e:
        calculator.run_custom_md(custom_parameters, return_last_atoms=True, only_prepare=True)
    assert 'Refusing to only prepare' in str(e)


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_run_custom_md_with_discarded_result(PbTe, calculator, custom_parameters):
    """
    Tests that running without returning last atoms
    is prevented (because all results would be gone).
    """
    PbTe.calc = calculator
    with pytest.raises(ValueError) as e:
        calculator.run_custom_md(custom_parameters, return_last_atoms=False)
    assert 'Refusing to run in temporary directory' in str(e)


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_run_custom_md_with_bad_input(PbTe, calculator, custom_parameters):
    """
    Tests that running without returning last atoms
    is prevented (because all results would be gone).
    """
    PbTe.calc = calculator
    custom_parameters[0] = ('bad', 'input')
    with pytest.raises(RuntimeError) as e:
        calculator.run_custom_md(custom_parameters, return_last_atoms=True)
    assert 'Calculator "gpunep" failed with command' in str(e)


@pytest.mark.parametrize('calculator',
                         [(3, True),
                          (3, False)], indirect=['calculator'],
                         )
def test_write_input_tmpdir(PbTe, calculator, single_point_parameters):
    """Tests that function controlling writing of input files works."""
    calculator.write_input(PbTe)
    dirs = os.listdir(calculator.directory)
    assert 'model.xyz' in dirs
    assert 'run.in' in dirs

    # Check that content of run.in is proper single point calculation parameters
    assert runfile_matches_parameters(os.path.join(calculator.directory, 'run.in'),
                                      single_point_parameters,
                                      calculator.potential_filename)


@pytest.mark.parametrize('calculator',
                         [(3, True),
                          (3, False),
                          (2, False)], indirect=['calculator'],
                         )
def test_write_runfile(calculator, custom_parameters):
    """Tests that single point parameters input file is properly written."""
    calculator._write_runfile(custom_parameters)
    assert runfile_matches_parameters(os.path.join(calculator.directory, 'run.in'),
                                      custom_parameters,
                                      calculator.potential_filename)


@pytest.mark.parametrize('calculator',
                         [(3, True)], indirect=['calculator'],
                         )
def test_write_runfile_with_specified_potential(calculator, custom_parameters):
    """Tests that single point parameters input file is properly written."""
    custom_parameters = [('potential', 'nep.txt')] + custom_parameters
    calculator._write_runfile(custom_parameters)
    assert runfile_matches_parameters(os.path.join(calculator.directory, 'run.in'),
                                      custom_parameters,
                                      'nep.txt')


@pytest.mark.parametrize('calculator, target_energy, target_stress',
                         [((3, True), -4.8747868538,
                           [-0.17525766, -0.17440966, -0.17390886,
                            -7.08215934e-08, -5.95844148e-08, -5.78122793e-03])],
                         indirect=['calculator'],
                         )
def test_get_potential_energy_and_stresses_from_file(PbTe, custom_parameters, calculator,
                                                     target_energy, target_stress):
    """Tests reading of potential energy and stresses."""
    PbTe.calc = calculator
    calculator.run_custom_md(custom_parameters, return_last_atoms=False)
    energy, stress = calculator.get_potential_energy_and_stresses_from_file()
    assert abs(energy - target_energy) < 1e-5
    assert np.allclose(stress, target_stress)


@pytest.mark.parametrize('calculator',
                         [(3, True)],
                         indirect=['calculator'],
                         )
def test_fail_to_get_potential_energy_and_stresses_from_file(PbTe, calculator, tmpdir):
    """Tests failure when reading potential energy and stresses."""
    PbTe.calc = calculator
    p = tmpdir.join('thermo.out')
    p.write('NaN NaN NaN NaN NaN NaN\n')
    with pytest.raises(ValueError) as e:
        calculator.get_potential_energy_and_stresses_from_file()
    assert 'Failed to extract energy and/or stresses:' in str(e)


@pytest.mark.parametrize('calculator, target_forces',
                         [((3, True), [[3.97509e-01, 2.73986e-01, 6.98623e-06],
                                       [-3.97509e-01, -2.73986e-01, -6.98623e-06]])],
                         indirect=['calculator'],
                         )
def test_get_forces_from_file(PbTe, custom_parameters, calculator, target_forces):
    """Tests reading of potential energy and stresses."""
    PbTe.calc = calculator
    calculator.run_custom_md(custom_parameters, return_last_atoms=False)
    forces = calculator.get_forces_from_file()
    assert np.allclose(forces, target_forces, atol=1e-6)


@pytest.mark.parametrize('calculator, expected_energy',
                         [((3, True), -4.8564407825),
                          ((3, False), -4.8564407825),
                          ((2, False), 58.0)],
                         indirect=['calculator'],
                         )
def test_get_potential_energy(PbTe, calculator, expected_energy):
    """Tests standard way of calculating potential energy."""
    PbTe.calc = calculator
    e = PbTe.get_potential_energy()
    assert abs(e - expected_energy) < 1e-5


@pytest.mark.parametrize('calculator, expected_forces',
                         [((3, True), [[1.61652e-01, 1.08260e-01, 1.67526e-06],
                                       [-1.61652e-01, -1.08260e-01, -1.67526e-06]]),
                          ((3, False), [[1.61652e-01, 1.08260e-01, 1.67526e-06],
                                        [-1.61652e-01, -1.08260e-01, -1.67526e-06]]),
                          ((2, False), [[0, 0, 0], [0, 0, 0]])],
                         indirect=['calculator'],
                         )
def test_get_forces(PbTe, calculator, expected_forces):
    """Tests standard way of calculating forces."""
    PbTe.calc = calculator
    forces = PbTe.get_forces()
    assert np.allclose(forces, expected_forces, atol=1e-6)


@pytest.mark.parametrize('calculator, expected_stress',
                         [((3, True), [-0.15837644, -0.15825044, -0.15815363,
                                       -1.95632527e-08, 5.19175762e-08, -9.06732487e-04]),
                          ((3, False), [-0.15837644, -0.15825044, -0.15815363,
                                        -1.95632527e-08, 5.19175762e-08, -9.06732487e-04]),
                          ((2, False), [0, 0, 0, 0, 0, 0])],
                         indirect=['calculator'],
                         )
def test_get_stress(PbTe, calculator, expected_stress):
    """Tests standard way of calculating forces."""
    PbTe.calc = calculator
    stress = PbTe.get_stress()
    assert np.allclose(stress, expected_stress, atol=1e-6)


@pytest.mark.parametrize('calculator',
                         [(3, True)], indirect=['calculator'],
                         )
def test_read_results(PbTe, calculator, custom_parameters):
    """Tests read_results function."""
    PbTe.calc = calculator
    calculator.run_custom_md(custom_parameters, return_last_atoms=False)
    assert len(calculator.results) == 0
    calculator.read_results()
    assert 'energy' in calculator.results
    assert 'forces' in calculator.results
    assert 'stress' in calculator.results


@pytest.mark.parametrize('calculator',
                         [(3, True)], indirect=['calculator'],
                         )
def test_clean(PbTe, calculator, custom_parameters):
    """Tests that directories are properly deleted."""
    PbTe.calc = calculator
    calculator.run_custom_md(custom_parameters, return_last_atoms=True)
    assert os.path.exists(calculator.directory)
    calculator._clean()
    assert not os.path.exists(calculator.directory)


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_make_new_tmp_directory(PbTe, calculator):
    """Tests creation of new tmp directory"""
    PbTe.calc = calculator
    assert os.path.exists(os.path.join(os.path.abspath(calculator.directory),
                                       calculator._potential_path))
    assert os.path.exists(calculator.directory)
    current_tempdir = calculator.directory
    calculator._make_new_tmp_directory()
    assert os.path.exists(os.path.join(os.path.abspath(calculator.directory),
                                       calculator._potential_path))
    assert calculator.directory == current_tempdir
    calculator._directory = None
    calculator._make_new_tmp_directory()
    assert calculator.directory != current_tempdir
    assert os.path.exists(os.path.join(os.path.abspath(calculator.directory),
                                       calculator._potential_path))


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_set_atoms(PbTe, calculator):
    """Tests set_atoms function"""
    assert calculator.atoms is None
    calculator.set_atoms(PbTe)
    assert isinstance(calculator.atoms, Atoms)


@pytest.mark.parametrize('calculator',
                         [(3, False)], indirect=['calculator'],
                         )
def test_set_directory(calculator, tmpdir):
    """Tests set_directory function"""
    assert calculator._use_temporary_directory
    calculator.set_directory(tmpdir)
    assert calculator.directory == tmpdir
    assert not calculator._use_temporary_directory
    assert os.path.exists(os.path.join(os.path.abspath(tmpdir), calculator._potential_path))


@pytest.mark.parametrize('calculator, expected_energy',
                         [((3, True), -4.8564407825),
                          ((3, False), -4.8564407825),
                          ((2, False), 58)],
                         indirect=['calculator'],
                         )
def test_calculator_reuse(PbTe, calculator, expected_energy):
    """Test that calculator can be reused with different atoms object."""
    PbTe.calc = calculator
    e = PbTe.get_potential_energy()
    assert abs(e - expected_energy) < 1e-5
    new_structure = PbTe.repeat(2)
    new_structure.calc = calculator
    e = new_structure.get_potential_energy()
    assert abs(e - 8 * expected_energy) < 1e-5



import numpy as np
import pytest
from ase.build import bulk
from ase.calculators.emt import EMT
from calorine.tools import relax_structure


@pytest.fixture
def structure():
    structure = bulk('Ag', a=4, cubic=True)
    structure[0].position += np.array([0.03, 0.08, 0])
    structure.calc = EMT()
    return structure


@pytest.mark.parametrize('minimizer, constant_cell, constant_volume, cell', [
    (
        'bfgs',
        False,
        False,
        [[4.064344467646021, 0.00010293213753658188, 0],
         [0.00010293213753565035, 4.063399911533125, 0],
         [0, 0, 4.064486069840965]]
    ),
    (
        'bfgs',
        False,
        True,
        [[4.000296209078849, -0.00023363048485108715, 0],
         [-0.000233630484850003, 3.9992734453197674, 0],
         [0, 0, 4.000430459369042]]
    ),
    (
        'fire',
        True,
        False,  # Volume is fixed if cell is fixed
        [[4.0, 0.0, 0.0],
         [0.0, 4.0, 0.0],
         [0.0, 0.0, 4.0]]
    ),
    (
        'bfgs-scipy',
        False,
        True,
        [[4.000288933940351, 0.0004604843003583068, 0],
         [0.00046048430035830777, 3.9992704548874647, 0],
         [0, 0, 4.000440765444412]]
    ),
    (
        'gpmin',
        False,
        True,
        [[3.9994073553656007, -0.0006934672576998964, 0],
         [-0.0006934672576774285, 4.00125729962704, 0],
         [0, 0, 3.9993357618357335]]
    ),
])
def test_check_fmax(structure,
                    minimizer,
                    constant_cell,
                    constant_volume,
                    cell):
    """Tests that a structure is relaxed to a maximum force lower than the specified threshold."""
    pos_pre_relax = structure.get_positions()

    relax_structure(
        structure,
        fmax=0.02,
        steps=1000,
        minimizer=minimizer,
        constant_cell=constant_cell,
        constant_volume=constant_volume,
    )

    fmax_post_relax = np.max(np.linalg.norm(structure.get_forces(), axis=1))
    assert fmax_post_relax < 0.03
    assert not np.allclose(structure.positions, pos_pre_relax)  # Make sure that the atoms moved
    assert np.isclose(structure.get_volume(), np.linalg.det(cell), atol=1e-4)
    assert np.allclose(structure.get_cell(), cell, atol=1e-4)


@pytest.mark.parametrize('with_calculator, minimizer, error',
                         [
                             (False, 'fire', 'Structure has no attached calculator object'),
                             (True, 'invalid', 'Unknown minimizer: invalid'),
                          ]
                         )
def test_no_calculator(structure, with_calculator, minimizer, error):
    if not with_calculator:
        structure.calc = None
    with pytest.raises(ValueError) as e:
        relax_structure(structure, minimizer=minimizer)
    assert error in str(e)



import pytest
import subprocess
import tempfile
import numpy as np
from ase.build import bulk
from ase.calculators.emt import EMT
from ase.io import read
from ase.units import GPa
from os.path import join as join_path
from calorine.nep import read_loss, read_nepfile, read_structures, setup_training


@pytest.fixture
def nep_parameters():
    return dict(version=4,
                type=[2, 'Au', 'Cu'],
                cutoff=[4, 4],
                n_max=[2, 1],
                l_max=[4, 0],
                lambda_1=0.1,
                lambda_2=0.2,
                lambda_e=1,
                lambda_f=5,
                lambda_v=0.2,
                neuron=2,
                generation=1000)


@pytest.fixture
def structures():
    structures = []

    structure = bulk('AuCu', crystalstructure='rocksalt', a=7.0)
    structure[0].z += 0.1
    structure.calc = EMT()
    structures.append(structure)

    structure = bulk('AuCu', crystalstructure='rocksalt', a=7.0)
    structure[0].z += 0.15
    structure.calc = EMT()
    structures.append(structure)

    return structures


@pytest.fixture
def setup_parameters():
    return dict(train_fraction=0.5,
                seed=42,
                n_splits=None,
                overwrite=True)


def test_nep_execution(nep_parameters, structures, setup_parameters):
    """This function tests running the nep executable with input generated
    by calorine. Since running the nep binary takes a couple of
    seconds to produce parseable output. The function contains
    multiple tests (rather than splitting them up in several different
    functions.)

    Note that this test requires the test runner to have a compatible GPU.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        setup_parameters['rootdir'] = tmpdirname
        setup_training(nep_parameters, structures, **setup_parameters)

        path = join_path(setup_parameters['rootdir'], 'nepmodel_full')
        result = subprocess.run(['nep'], capture_output=True, cwd=path)
        assert 'Finished running nep.' in str(result.stdout)

        # check that read_loss runs
        _ = read_loss(join_path(path, 'loss.out'))

        # check that read_nepfile runs
        _ = read_nepfile(join_path(path, 'nep.in'))

        # test read_structures
        train_structures, test_structures = read_structures(path)
        train_structures_ref = read(join_path(path, 'train.xyz'), ':')
        test_structures_ref = read(join_path(path, 'test.xyz'), ':')

        assert len(train_structures) == len(train_structures_ref), \
            'number of training structures inconsistent'
        assert len(test_structures) == len(test_structures_ref), \
            'number of test structures inconsistent'

        for sref, s in zip(structures, train_structures + test_structures):
            assert len(sref) == len(s)
            assert np.allclose(sref.positions, s.positions, atol=1e-6)
            assert np.allclose(sref.get_forces(), s.get_forces(), atol=1e-6)
            assert np.allclose(sref.cell, s.cell, atol=1e-6)
            assert np.all(sref.symbols == s.symbols)

            for key in ['energy_predicted', 'energy_target',
                        'force_predicted', 'force_target',
                        'virial_predicted', 'virial_target',
                        'stress_predicted', 'stress_target']:
                assert key in s.info

            assert np.allclose(sref.get_forces(), s.info['force_target'], atol=1e-6)
            assert np.isclose(sref.get_potential_energy() / len(sref),
                              s.info['energy_target'], atol=1e-6)
            stress = -sref.get_stress() / GPa
            assert np.allclose(stress, s.info['stress_target'], atol=1e-6)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import sys
import sphinx_rtd_theme

sys.path.insert(0, os.path.abspath('../../'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.ifconfig',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.napoleon',
    'sphinx.ext.graphviz',
    'sphinx_autodoc_typehints',
    'sphinx_sitemap',
    'cloud_sptheme.ext.table_styling',
    'nbsphinx']

graphviz_output_format = 'svg'
templates_path = ['_templates']
default_role = 'code'
source_suffix = '.rst'
master_doc = 'index'
pygments_style = 'sphinx'
todo_include_todos = True

# Collect basic information from main module
with open('../calorine/__init__.py') as fd:
    lines = '\n'.join(fd.readlines())
version = ''
if len(version) == 0:
    version = re.search("__version__ = '(.*)'", lines).group(1)
release = ''
copyright = re.search("__copyright__ = '(.*)'", lines).group(1)
project = re.search("__project__ = '(.*)'", lines).group(1)
author = re.search("__maintainer__ = '(.*)'", lines).group(1)

site_url = 'https://calorine.materialsmodeling.org/'
html_css_files = ['custom.css']
html_logo = '_static/logo.png'
html_favicon = '_static/logo.ico'
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = ['_static']
html_theme_options = {'display_version': True}
html_context = {
    'current_version': version,
    'versions':
        [('latest stable release',
          '{}'.format(site_url)),
         ('development version',
          '{}/dev'.format(site_url))]}
htmlhelp_basename = 'calorinedoc'
intersphinx_mapping = \
    {'ase':     ('https://wiki.fysik.dtu.dk/ase', None),
     'numpy':   ('https://numpy.org/doc/stable/', None),
     'h5py':    ('http://docs.h5py.org/en/latest/', None),
     'scipy':   ('https://scipy.github.io/devdocs/', None),
     'sklearn': ('https://scikit-learn.org/stable', None)}

# Settings for nbsphinx
nbsphinx_execute = 'never'

# Options for LaTeX output
_PREAMBLE = r"""
\usepackage{amsmath,amssymb}
\renewcommand{\vec}[1]{\boldsymbol{#1}}
\DeclareMathOperator*{\argmin}{\arg\!\min}
\DeclareMathOperator{\argmin}{\arg\!\min}
"""

latex_elements = {
    'preamble': _PREAMBLE,
}
latex_documents = [
    (master_doc, 'calorine.tex', 'calorine Documentation',
     'The calorine developer team', 'manual'),
]


# Options for manual page output
man_pages = [
    (master_doc, 'calorine', 'calorine Documentation',
     [author], 1)
]


# Options for Texinfo output
texinfo_documents = [
    (master_doc, 'calorine', 'calorine Documentation',
     author, 'calorine', 'Strong coupling calculator',
     'Miscellaneous'),
]



import tempfile

import numpy as np
import pytest
from ase import Atoms
from ase.build import bulk
from ase.io import read, write
from calorine.gpumd import (read_thermo,
                            read_kappa, read_hac,
                            read_xyz, write_xyz)


# --- read_kappa ---
def test_read_kappa():
    """Reads a kappa.out file"""
    kappa = read_kappa('tests/example_files/kappa.out')
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_single_row(tmpdir):
    """Reads a single row kappa.out file"""
    p = tmpdir.join('kappa.out')
    p.write('0 1 2 3 4\n')
    kappa = read_kappa(str(p))
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_nan():
    """Reads a kappa.out file filed with nans. Should this raise a warning?"""
    kappa = read_kappa('tests/example_files/kappa_nan.out')
    columns_check = kappa.columns == ['kx_in', 'kx_out', 'ky_in', 'ky_out',
                                      'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(kappa.index[0], int)
    assert isinstance(kappa['kx_in'][0], np.float64)
    assert kappa.index[0] == 0


def test_read_kappa_malformed_file(tmpdir):
    """Tries to read a malformed kappa.out file"""
    p = tmpdir.join('kappa_invalid.out')
    p.write('0 0 0 0 0 0 0 0 0 0 \n')
    with pytest.raises(ValueError) as e:
        read_kappa(str(p))
    assert 'Input file contains 10 data columns. Expected 5 columns.' in str(e)


# --- read_hac ---
def test_read_hac():
    """Reads a hac.out file"""
    hac = read_hac('tests/example_files/hac.out')
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_single_row(tmpdir):
    """Reads a single row hac.out file"""
    p = tmpdir.join('hac.out')
    p.write('0 1 2 3 4 5 6 7 8 9 10\n')
    hac = read_hac(str(p))
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_nan():
    """Reads a hac.out file filed with nans. Should this raise a warning?"""
    hac = read_hac('tests/example_files/hac_nan.out')
    columns_check = hac.columns == ['time', 'kx_out', 'ky_out',
                                    'kz_tot', 'kx_tot', 'ky_tot']
    assert columns_check.all()
    assert isinstance(hac.index[0], int)
    assert isinstance(hac['time'][0], np.float64)
    assert hac.index[0] == 0


def test_read_hac_malformed_file(tmpdir):
    """Tries to read a malformed hac.out file"""
    p = tmpdir.join('hac_invalid.out')
    p.write('0 0 0 \n')
    with pytest.raises(ValueError) as e:
        read_hac(str(p))
    assert 'Input file contains 3 data columns. Expected 11 columns.' in str(e)


# --- read_thermo ---
def test_read_thermo_orthorhombic():
    """Reads a thermo.out file with an orthorhombic structure"""
    thermo = read_thermo('tests/example_files/thermo_ortho_v3.2.out')
    columns_check = thermo.columns == ['temperature', 'kinetic_energy', 'potential_energy',
                                       'stress_xx', 'stress_yy', 'stress_zz',
                                       'cell_xx', 'cell_yy', 'cell_zz']
    assert columns_check.all()


def test_read_thermo_triclinic():
    """Reads a thermo.out file with an triclinic structure"""
    thermo = read_thermo('tests/example_files/thermo_tri_v3.2.out')
    columns_check = thermo.columns == ['temperature', 'kinetic_energy', 'potential_energy',
                                       'stress_xx', 'stress_yy', 'stress_zz',
                                       'cell_xx', 'cell_xy', 'cell_xz',
                                       'cell_yx', 'cell_yy', 'cell_yz',
                                       'cell_zx', 'cell_zy', 'cell_zz'
                                       ]
    assert columns_check.all()


@pytest.mark.parametrize('test_input',
                         [
                             # 9 columns --> orthorhombic cell pre GPUMD v3.3.1
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'cell_xx', 'cell_yy', 'cell_zz']),
                             # 15 columns --> triclinc cell pre GPUMD v3.3.1
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'cell_xx', 'cell_xy', 'cell_xz',
                               'cell_yx', 'cell_yy', 'cell_yz',
                               'cell_zx', 'cell_zy', 'cell_zz']),
                             # 12 columns --> orthorhombic cell GPUMD v3.3.1 forward
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'stress_yz', 'stress_xz', 'stress_xy',
                               'cell_xx', 'cell_yy', 'cell_zz']),
                             # 18 columns --> triclinic cell GPUMD v3.3.1 forward
                             (['temperature', 'kinetic_energy', 'potential_energy',
                               'stress_xx', 'stress_yy', 'stress_zz',
                               'stress_yz', 'stress_xz', 'stress_xy',
                               'cell_xx', 'cell_xy', 'cell_xz',
                               'cell_yx', 'cell_yy', 'cell_yz',
                               'cell_zx', 'cell_zy', 'cell_zz'])
                         ])
def test_read_thermo_pass(test_input):
    """Reads dummy thermo.out files and checks that the correct columns are being returned"""
    s = ' '.join(map(str, range(len(test_input)))) + '\n'
    tmpfile = tempfile.NamedTemporaryFile()
    with open(tmpfile.name, 'w') as f:
        for _ in range(10):
            f.write(s)
    with open(tmpfile.name, 'r') as f:
        thermo = read_thermo(f.name)
    tmpfile.close()
    columns_check = thermo.columns == test_input
    assert columns_check.all()


def test_read_thermo_fail():
    """Checks that ValueError is raised if the number of columns is incorrect"""
    for n in range(1, 20):
        s = ' '.join(map(str, range(n))) + '\n'
        tmpfile = tempfile.NamedTemporaryFile()
        with open(tmpfile.name, 'w') as f:
            for _ in range(10):
                f.write(s)
        with open(tmpfile.name, 'r') as f:
            if n in [9, 15, 12, 18]:
                _ = read_thermo(f.name)
            else:
                with pytest.raises(ValueError) as e:
                    read_thermo(f.name)
                assert 'Expected 9, 12, 15 or 18 columns.' in str(e)
        tmpfile.close()


def test_read_thermo_malformed_file(tmpdir):
    """Tries to read a malformed thermo.out file"""
    p = tmpdir.join('thermo_invalid.out')
    p.write('NaN NaN NaN NaN NaN NaN\n')
    with pytest.raises(ValueError) as e:
        read_thermo(str(p))
    assert 'Input file contains 6 data columns. Expected 9, 12, 15 or 18 columns.' in str(e)


# --- read_xyz and write_xyz ---
def test_write_read_xyz(tmpdir):
    """Writes and reads xyz files"""

    f = str(tmpdir.join('atoms.xyz'))
    structure_orig = bulk('C').repeat(3)
    write_xyz(f, structure_orig)
    structure_read = read_xyz(f)
    assert len(structure_orig) == len(structure_read)
    assert np.allclose(structure_orig.cell, structure_read.cell)
    assert np.allclose(structure_orig.positions, structure_read.positions)

    write_xyz(f, structure_orig)
    structure_read = read_xyz(f)
    velocities = structure_read.get_velocities()
    assert np.max(np.abs(velocities)) < 1e-6


def test_write_read_xyz_has_velocity(tmpdir):
    """Writes and reads an orthorhombic structure with velocity"""
    f = str(tmpdir.join('atoms.xyz'))
    structure = Atoms('CCC', positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2)], cell=[1, 2, 3])
    vel = [[0, 0, 100], [0, 0, 150], [0, 0, 200]]
    structure.set_velocities(vel)
    write_xyz(f, structure)
    structure_read = read_xyz(f)
    velocities = structure_read.get_velocities()
    assert np.allclose(velocities, vel)


# def test_write_read_xyz_groupings(tmpdir):
    """Writes and reads a structure with groups"""
    f = str(tmpdir.join('atoms.xyz'))
    structure = Atoms('CCCC',
                      positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2), (0, 0, 3.3)],
                      cell=[1, 2, 3])
    write_xyz(f, structure, groupings=[[[0, 1], [2, 3]], [[0], [1, 2, 3]]])
    structure_read = read_xyz(f)
    assert len(structure_read.get_array('group')[0]) == 2


def test_write_xyz_invalid_groupings(tmpdir):
    """Tries to write with invalid groups"""
    f = tmpdir.join('atoms.xyz')
    structure = Atoms('CCC', positions=[(0, 0, 0), (0, 0, 1.1), (0, 0, 2.2)], cell=[1, 2, 3])
    with pytest.raises(ValueError) as e:
        # Too many groupings
        write_xyz(f, structure, groupings=[[[0, 1], [2]], [
                  [0, 1], [2]], [[0, 1], [2]], [[0, 1], [2]]])
    assert 'There can be no more than 3 grouping methods!' in str(e)
    with pytest.raises(ValueError) as e:
        # Number of atoms do not add up to the total
        write_xyz(f, structure, groupings=[[[0, 1], [1, 2]]])
    assert 'method 0 are not compatible with the input structure!' in str(e)


# --- Check dump.xyz ---
# Not a unit test, but a good FYI test for our sake
def test_ase_correctly_parses_dump(tmpdir):
    """Check that ASE can correctly read dump files"""
    dump_file = 'tests/example_files/md_no_velocities_or_forces/dump.xyz'
    snapshots = read(dump_file, index=':')
    traj = tmpdir.join('lmao.traj')
    write(f'{traj}', snapshots)  # Make sure that writing works without crashing
    assert all(snapshots[-1][-1].position == [3.10105816, 2.94360168, 2.68268773])


def test_ase_correctly_parses_dump_forces_and_velocities(tmpdir):
    """Check that ASE can correctly read dump files with forces and velocities"""
    dump_file = 'tests/example_files/md_velocities_and_forces/dump.xyz'
    snapshots = read(dump_file, index=':')
    traj = tmpdir.join('lmao.traj')
    write(f'{traj}', snapshots)  # Make sure that writing works without crashing
    assert all(snapshots[-1][-1].position == [3.24197501, 2.84584855, 2.89100032])
    assert all(snapshots[-1].get_array('vel')[-1] == [-0.00033243, 0.00061451, 0.00044743])
    assert all(snapshots[-1].get_array('forces')[-1] == [-0.96089202, -0.01649800, -0.11494368])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import sys

import setuptools
from setuptools import find_packages, setup
from setuptools.command.build_ext import build_ext
from pybind11.setup_helpers import Pybind11Extension


nepy_cpp_module = Pybind11Extension(
    '_nepy',
    ['src/nepy/nepy.cpp'],
    language='c++')


# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """
    Returns a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile
    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """
    Returns the -std=c++[11/14] compiler flag.

    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError('Unsupported compiler -- at least C++11 support '
                           'is needed!')


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""
    c_opts = {
        'msvc': ['/EHsc'],
        'unix': [],
    }

    if sys.platform == 'darwin':
        c_opts['unix'] += ['-mmacosx-version-min=10.7']

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if sys.platform == 'darwin':
            if has_flag(self.compiler, '-stdlib=libc++'):
                opts.append('-stdlib=libc++')
        if ct == 'unix':
            opts.append("-DVERSION_INFO='{}'"
                        .format(self.distribution.get_version()))
            opts.append(cpp_flag(self.compiler))
            if has_flag(self.compiler, '-fvisibility=hidden'):
                opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append("/DVERSION_INFO=\\'{}\\'"
                        .format(self.distribution.get_version()))
        opts.append('-O3')
        opts.append('-shared')
        opts.append('-fPIC')
        for ext in self.extensions:
            ext.extra_compile_args = opts
            ext.extra_link_args = opts
        build_ext.build_extensions(self)


with open('README.rst', encoding='utf-8') as fd:
    long_description = fd.read()

with open('calorine/__init__.py', encoding='utf-8') as fd:
    try:
        lines = ''
        for item in fd.readlines():
            item = item
            lines += item + '\n'
    except Exception as exc:
        raise Exception('Caught exception {}'.format(exc))


version = re.search("__version__ = '(.*)'", lines).group(1)
maintainer = re.search("__maintainer__ = '(.*)'", lines).group(1)
description = re.search("__description__ = '(.*)'", lines).group(1)
url = re.search("__url__ = '(.*)'", lines).group(1)
license = re.search("__license__ = '(.*)'", lines).group(1)

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: {}'.format(license),
    'Topic :: Scientific/Engineering :: Physics']


if __name__ == '__main__':

    setup(
        name='calorine',
        version=version,
        author='calorine developer group',
        description=description,
        long_description=long_description,
        ext_modules=[nepy_cpp_module],
        install_requires=['ase',
                          'matplotlib',
                          'numpy',
                          'pandas>=0.23',
                          'pybind11'],
        packages=find_packages(),
        python_requires='>=3.8',
        cmdclass={'build_ext': BuildExt},
        zip_safe=False,
        classifiers=classifiers,
        license=license,
        url=url,
    )



from typing import Any, Dict
from ase import Atoms
from ase.constraints import ExpCellFilter
from ase.optimize import BFGS, FIRE, GPMin
from ase.optimize.sciopt import SciPyFminBFGS


def relax_structure(structure: Atoms,
                    fmax: float = 0.001,
                    steps: int = 500,
                    minimizer: str = 'bfgs',
                    constant_cell: bool = False,
                    constant_volume: bool = False,
                    kwargs: Dict[str, Any] = dict()) -> None:
    """Relaxes the given structure.

    Parameters
    ----------
    structure
        atomic configuration to relax
    fmax
        if the absolute force for all atoms falls below this value the relaxation is stopped
    steps
        maximum number of relaxation steps the minimizer is allowed to take
    minimizer
        minimizer to use; possible values: 'bfgs', 'cg', 'fire', 'gpmin', 'bfgs-scipy'
    constant_cell
        if True do not relax the cell or the volume
    constant_volume
        if True relax the cell shape but keep the volume constant
    kwargs
        keyword arguments to be handed over to the minimizer; possible arguments can be found
        in the [ASE documentation](https://wiki.fysik.dtu.dk/ase/ase/optimize.html)
    """
    if structure.calc is None:
        raise ValueError('Structure has no attached calculator object')
    if constant_cell:
        ucf = structure
    else:
        ucf = ExpCellFilter(structure, constant_volume=constant_volume)
    loc_kwargs = kwargs.copy()
    loc_kwargs['logfile'] = loc_kwargs.get('logfile', None)
    if minimizer == 'bfgs':
        dyn = BFGS(ucf, **loc_kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'bfgs-scipy':
        dyn = SciPyFminBFGS(ucf, **loc_kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'fire':
        dyn = FIRE(ucf, **loc_kwargs)
        dyn.run(fmax=fmax, steps=steps)
    elif minimizer == 'gpmin':
        dyn = GPMin(ucf, **loc_kwargs)
        dyn.run(fmax=fmax, steps=steps)
    else:
        raise ValueError(f'Unknown minimizer: {minimizer}')



import contextlib
import numpy as np
import pytest
from ase import Atoms
from ase.build import bulk
from ase.stress import full_3x3_to_voigt_6_stress
from ase.db import connect
from calorine.calculators import CPUNEP, GPUNEP


# Use same volume as in PbTe_2atom.in, C_2atom.in etc.
vacuum_cell = ([100, 0, 0], [0, 100, 0], [0, 0, 100])
PbTe = Atoms('TePb', positions=[(0, 0, 0), (0, 0.0, 1.1)],
             cell=vacuum_cell)
C = Atoms('C', positions=[(0, 0, 0)],
          cell=vacuum_cell)
CO = Atoms('CO', positions=[(0, 0, 0), (0, 0.0, 1.1)],
           cell=vacuum_cell)
CC = Atoms('CC', positions=[(0, 0, 0), (0, 0.0, 1.1)],
           cell=vacuum_cell)
CON = Atoms('CON', positions=[(0, 0, 0), (0, 0.0, 1.1), (0, 0.0, 2.2)],
            cell=vacuum_cell)
PbTeBulk = bulk('PbTe', crystalstructure='rocksalt', a=4)
PbTeBulk[0].position += np.array([0.03, 0.02, 0])


@pytest.fixture
def NEP3CPUNEP():
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    return calc


def get_expected_forces(path):
    return np.loadtxt(path)


def get_expected_stress(path, voigt=True):
    volume = vacuum_cell[0][0] * vacuum_cell[1][1] * vacuum_cell[2][2]
    stress = (np.sum(-np.loadtxt(path), axis=0) / volume).reshape((3, 3))
    if voigt:
        return full_3x3_to_voigt_6_stress(stress)
    return stress


# --- get_potential_forces_and_virials ---
def test_get_potential_forces_and_stress_NEP3():
    """NEP3 model supplied. Compares results to output from `nep_cpu`"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    PbTe.calc = calc
    energy = PbTe.get_potential_energy()
    forces = PbTe.get_forces()
    stress = PbTe.get_stress()

    expected_forces = get_expected_forces(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_force.out')
    expected_stress = get_expected_stress(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_virial.out')

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert stress.shape == (6,)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_set_atoms_constructor():
    """Set atoms directly when creating the calculator"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    CPUNEP(nep3, atoms=PbTe)
    energy = PbTe.get_potential_energy()
    forces = PbTe.get_forces()
    stress = PbTe.get_stress()

    expected_forces = get_expected_forces(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_force.out')
    expected_stress = get_expected_stress(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_virial.out')

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert stress.shape == (6,)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_set_atoms_calculate():
    """Set atoms directly when calling calculate on the calculator"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    calc.calculate(atoms=PbTe)
    results = calc.results
    energy = results['energy']
    forces, stress = results['forces'], results['stress']

    expected_forces = get_expected_forces(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_force.out')
    expected_stress = get_expected_stress(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_virial.out', voigt=True)

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert stress.shape == (6, )
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_NEP3_debug(tmpdir):
    """Compares result with debug flag enabled."""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3, debug=True)
    p = tmpdir.join('nep_out.tmp')
    with open(p, 'w') as f:
        with contextlib.redirect_stdout(f):
            with contextlib.redirect_stderr(f):
                PbTe.calc = calc
                PbTe.get_potential_energy()
                PbTe.get_forces()
                PbTe.get_stress()
    with open(p, 'r') as f:
        lines = p.readlines()
        assert lines[0] == 'Use the NEP3 potential with 2 atom types.\n'
        assert len(lines) == 16


def test_get_tress_non_voigt_NEP3():
    """NEP3 model supplied. Compares results to output from `nep_cpu`"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    PbTe.calc = calc
    stress = PbTe.get_stress(voigt=False)
    expected_stress = get_expected_stress(
        'tests/example_output/nep3_v3.3.1_PbTe_Fan22_PbTe_2atom_virial.out', voigt=False)

    assert stress.shape == (3, 3)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_dummy_NEP2():
    """Dummy NEP2 model supplied. Compares results to output from `nep_cpu` for another system"""
    nep2 = 'tests/nep_models/CO_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CO.calc = calc
    energy = CO.get_potential_energy()
    forces = CO.get_forces()
    stress = CO.get_stress()

    expected_forces = get_expected_forces(
        'tests/example_output/CO_NEP2_dummy_CO_2atom_force.out')
    expected_stress = get_expected_stress(
        'tests/example_output/CO_NEP2_dummy_CO_2atom_virial.out')

    assert energy.shape == ()
    assert forces.shape == (2, 3)
    assert stress.shape == (6,)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_several_different_species():
    """Check that forces and stress are correct for a CON system.
    Note that these forces/stress should be exactly zero for this system
    since the NEP2 dummy potential treats all atom species as identical atm."""
    nep2 = 'tests/nep_models/CON_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CON.calc = calc
    energy = CON.get_potential_energy()
    forces = CON.get_forces()
    stress = CON.get_stress()

    expected_forces = get_expected_forces(
        'tests/example_output/CON_NEP2_dummy_CON_3atom_force.out')
    expected_stress = get_expected_stress(
        'tests/example_output/CON_NEP2_dummy_CON_3atom_virial.out')

    assert energy.shape == ()
    assert forces.shape == (3, 3)
    assert stress.shape == (6,)
    assert np.allclose(forces[0, :], -forces[1, :], atol=1e-12, rtol=0)  # Newton III
    assert np.allclose(forces, expected_forces, atol=1e-12, rtol=0)
    assert np.allclose(forces, np.zeros((3, 3)), atol=1e-12, rtol=0)
    assert np.allclose(stress, expected_stress, atol=1e-12, rtol=0)
    assert np.allclose(stress, np.zeros((6,)), atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_dummy_NEP2_independent_of_species():
    """Dummy NEP2 energies, forces and stress should be independent of atom species"""
    nep2 = 'tests/nep_models/PbTe_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    PbTe.calc = calc
    PbTe_energy = PbTe.get_potential_energy()
    PbTe_forces = PbTe.get_forces()
    PbTe_stress = PbTe.get_stress()

    nep2 = 'tests/nep_models/CO_NEP2_dummy.txt'
    calc = CPUNEP(nep2)
    CO.calc = calc
    CO_energy = CO.get_potential_energy()
    CO_forces = CO.get_forces()
    CO_stress = CO.get_stress()

    expected_forces_CO = get_expected_forces(
        'tests/example_output/CO_NEP2_dummy_CO_2atom_force.out')
    expected_stress_CO = get_expected_stress(
        'tests/example_output/CO_NEP2_dummy_CO_2atom_virial.out')

    assert np.allclose(PbTe_energy, CO_energy, atol=1e-12, rtol=0)
    assert np.allclose(PbTe_forces, CO_forces, atol=1e-12, rtol=0)
    assert np.allclose(PbTe_stress, CO_stress, atol=1e-12, rtol=0)
    assert np.allclose(CO_forces, expected_forces_CO, atol=1e-12, rtol=0)
    assert np.allclose(CO_stress, expected_stress_CO, atol=1e-12, rtol=0)


def test_get_potential_forces_and_stress_update_positions():
    """Update the positions and make sure that the energies, forces and stress are also updated"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()
    stress_initial = copy.get_stress()

    # Move atoms slightly
    copy.set_positions([[0, 0, 0], [0, 0, 2.2]])
    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()
    stress_after = copy.get_stress()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    diff_stress = stress_initial - stress_after

    assert np.isclose(diff_energy, 1.80751674, atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_force, [[0, 0, 4.65672972], [0, 0, -4.65672972]], atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_stress, [0, 0, 7.13057884e-06, 0, 0, 0], atol=1e-12, rtol=1e-6)


def test_get_potential_forces_and_stress_update_cell():
    """Update the cell and make sure that the energies, forces and stress are still the same"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()
    stress_initial = copy.get_stress()

    # Change box
    new_cell = ([20, 0, 0], [0, 20, 0], [0, 0, 20])
    copy.set_cell(new_cell, scale_atoms=False)
    volume_factor = vacuum_cell[0][0]**3 / (new_cell[0][0]**3)
    copy.center()

    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()
    stress_after = copy.get_stress()

    assert np.isclose(energy_initial, energy_after, atol=1e-12, rtol=1e-6)
    assert np.allclose(forces_initial, forces_after, atol=1e-12, rtol=1e-6)
    assert np.allclose(stress_initial, stress_after / volume_factor, atol=1e-12, rtol=1e-6)


def test_get_potential_forces_and_stress_update_numbers():
    """Update the atom numbers (species) and make sure that the
    energies, forces and stress are also updated.
    """
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()
    stress_initial = copy.get_stress()

    # Change atomic numbers
    copy.set_atomic_numbers([82, 82])  # Pb_2

    energy_after = copy.get_potential_energy()
    forces_after = copy.get_forces()
    stress_after = copy.get_stress()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    diff_stress = stress_initial - stress_after
    assert np.isclose(diff_energy, 1.86577361, atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_force, [[0, 0, 1.07038059], [0, 0, -1.07038059]], atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_stress, [0, 0, 1.17741865e-06, 0, 0, 0], atol=1e-12, rtol=1e-6)


def test_reset_calculator_on_atoms_change():
    """Reset the calculator when changing the system.
    """
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)

    copy = PbTe.copy()
    copy.calc = calc
    energy_initial = copy.get_potential_energy()
    forces_initial = copy.get_forces()
    stress_initial = copy.get_stress()

    # Copy system
    atoms_copy = copy.copy()
    original_cell = copy.cell.copy()
    atoms_copy.calc = calc

    assert calc.results == {}
    assert calc.nepy is None

    # Scale cell
    atoms_copy.set_cell(1.1*original_cell, scale_atoms=True)
    energy_after = atoms_copy.get_potential_energy()
    forces_after = atoms_copy.get_forces()
    stress_after = atoms_copy.get_stress()

    diff_energy = np.abs(energy_after - energy_initial)
    diff_force = forces_initial - forces_after
    diff_stress = stress_initial - stress_after
    assert np.isclose(diff_energy, 0.28572432, atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_force, [[0, 0, 0.27278336], [0, 0, -0.27278336]], atol=1e-12, rtol=1e-6)
    assert np.allclose(diff_stress, [0, 0, 7.88470474e-07, 0, 0, 0], atol=1e-12, rtol=1e-6)


def test_get_potential_and_forces_no_cell():
    """Should raise error if no cell is supplied"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    atoms = Atoms('C', positions=[(0, 0, 0)])
    atoms.calc = calc
    with pytest.raises(ValueError) as e:
        atoms.get_potential_energy()
    assert 'Atoms must have a defined cell.' in str(e)


def test_get_potential_and_forces_no_potential():
    """Tries to get potentials and forces without specifying potential"""
    with pytest.raises(FileNotFoundError)as e:
        CPUNEP('nep.txt')
    assert 'nep.txt does not exist.' in str(e)


def test_get_potential_and_forces_no_atoms():
    """Tries to get potential and forces without specifying atoms"""
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    calc = CPUNEP(nep3)
    with pytest.raises(ValueError)as e:
        calc.calculate()
    assert 'Atoms must be defined to get energies and forces.' in str(e)


def test_CPU_GPU_equivalent():
    """Assert that the CPU and GPU implementation are equivalent
    """
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    # CPU
    PbTe.calc = CPUNEP(nep3)
    cpu_energy = PbTe.get_potential_energy()
    cpu_forces = PbTe.get_forces()
    cpu_stress = PbTe.get_stress()
    # GPU
    PbTe.calc = GPUNEP(nep3)
    gpu_energy = PbTe.get_potential_energy()
    gpu_forces = PbTe.get_forces()
    gpu_stress = PbTe.get_stress()
    assert np.isclose(cpu_energy, gpu_energy, atol=1e-12, rtol=1e-5)
    # GPUMD forces are in single precision, meaning errors can add up to 1e-6
    assert np.allclose(cpu_forces, gpu_forces, atol=1e-12, rtol=1e-5)
    assert np.allclose(cpu_stress, gpu_stress, atol=1e-12, rtol=1e-5)


def test_CPU_GPU_equivalent_bulk():
    """Assert that the CPU and GPU implementation are equivalent for a bulk PbTE system
    """
    nep3 = 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    print(PbTeBulk.get_volume())
    # CPU
    PbTeBulk.calc = CPUNEP(nep3)
    cpu_energy = PbTeBulk.get_potential_energy()
    cpu_forces = PbTeBulk.get_forces()
    cpu_stress = PbTeBulk.get_stress(voigt=False)
    # GPU
    PbTeBulk.calc = GPUNEP(nep3)
    gpu_energy = PbTeBulk.get_potential_energy()
    gpu_forces = PbTeBulk.get_forces()
    gpu_stress = PbTeBulk.get_stress(voigt=False)
    assert np.isclose(cpu_energy, gpu_energy, atol=1e-5, rtol=1e-5)
    # GPUMD forces are in single precision, meaning errors can add up to 1e-6
    assert np.allclose(cpu_forces, gpu_forces, atol=1e-5, rtol=1e-5)
    assert np.allclose(cpu_stress, gpu_stress, atol=1e-5, rtol=1e-5)


def test_cpunep_readwrite_dict(NEP3CPUNEP):
    dict = NEP3CPUNEP.todict()
    assert dict['potential_filename'] == 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt'
    restored_calc = CPUNEP(**dict)
    all_keys = dict.keys()
    assert all([getattr(restored_calc, key) == getattr(NEP3CPUNEP, key) for key in all_keys])


def test_cpunep_tostr(NEP3CPUNEP):
    s = str(NEP3CPUNEP)
    assert 'tests/nep_models/nep3_v3.3.1_PbTe_Fan22.txt' in s


def test_cpunep_readwrite_db(tmpdir, NEP3CPUNEP):
    PbTeBulk.calc = NEP3CPUNEP
    with connect(f'{tmpdir}/db.db') as db:
        db.write(PbTeBulk)
    atoms = None
    with connect(f'{tmpdir}/db.db') as db:
        for row in db.select():
            atoms = row.toatoms()
            atoms.calc = CPUNEP(**row.calculator_parameters)
    assert atoms == PbTeBulk
    assert atoms.calc is not None
    all_keys = ['potential_filename']
    assert all([getattr(atoms.calc, key) == getattr(NEP3CPUNEP, key) for key in all_keys])



from .io import (read_hac,
                 read_kappa,
                 read_thermo,
                 read_xyz,
                 write_xyz)

__all__ = ['read_hac',
           'read_kappa',
           'read_thermo',
           'read_xyz',
           'write_xyz']
